#include "DDiscretization.h"
#include <cmath>
#include <algorithm>

using namespace vii;

Discretization::Discretization()
{
    nMarkers = 0;
    nElm = 0;
    prevnMarkers = 0;
}

Discretization::~Discretization() {}

void Discretization::setGlob(std::vector<double> &_x, std::vector<double> &_y, std::vector<double> &_z)
{
    if (_x.size() != _y.size() || _y.size() != _z.size() || _x.size() != _z.size())
        throw std::runtime_error("vii::Discretization Wrong mesh sizes.\n");

    nMarkers = _x.size();
    x.resize(nMarkers, 0.);
    xoc.resize(nMarkers, 0.);
    y.resize(nMarkers, 0.);
    z.resize(nMarkers, 0.);
    nElm = nMarkers - 1;
    x = _x;
    y = _y;
    z = _z;

    computeBLcoord();
    computeOCcoord();
}

/**
 * @brief Compute x/chord coordinate.
*/
void Discretization::computeOCcoord()
{
    // Find min and max of array x
    auto min_it = std::min_element(x.begin(), x.end());
    double xMin = *min_it;
    auto max_it = std::max_element(x.begin(), x.end());
    double xMax = *max_it;

    // Compute xoc
    chord = xMax - xMin;
    if (chord <= 0)
        throw;
    for (size_t iPoint = 0; iPoint < x.size(); ++iPoint)
        xoc[iPoint] = (x[iPoint] - xMin) / chord;
}

/**
 * @brief Compute nodes coordinates in the local frame of reference.
 */
void Discretization::computeBLcoord()
{
    loc.resize(nMarkers, 0.);
    alpha.resize(nElm, 0.);
    dx.resize(nElm, 0.);

    for (size_t iPoint = 0; iPoint < nElm; ++iPoint)
    {
        alpha[iPoint] = std::atan2(y[iPoint + 1] - y[iPoint], x[iPoint + 1] - x[iPoint]);
        // dx = sqrt(delta x ^2 + delta y ^2).
        dx[iPoint] = std::sqrt((x[iPoint + 1] - x[iPoint]) * (x[iPoint + 1] - x[iPoint]) + (y[iPoint + 1] - y[iPoint]) * (y[iPoint + 1] - y[iPoint]));
        loc[iPoint + 1] = loc[iPoint] + dx[iPoint];
    }
}

void Discretization::setExt(std::vector<double> extM)
{
    ext.resize(extM.size(), 0.);
    ext = extM;
};
